@@ -161,7 +161,6 @@ static void brw_debug(ScrnInfoPtr scrn, char *when)
 	int i;
 	uint32_t v;
 
-	intel_sync(scrn);
 	ErrorF("brw_debug: %s\n", when);
 	for (i = 0; svg_ctl_bits[i].name; i++) {
 		OUTREG(BRW_SVG_CTL, svg_ctl_bits[i].svg_ctl);
@@ -779,18 +778,12 @@ i965_emit_video_setup(ScrnInfoPtr scrn, drm_intel_bo * bind_bo, int n_src_surf)
 	urb_cs_start = urb_sf_start + urb_sf_size;
 	urb_cs_size = URB_CS_ENTRIES * URB_CS_ENTRY_SIZE;
 
-	ATOMIC_BATCH(2);
 	OUT_BATCH(MI_FLUSH |
 		  MI_STATE_INSTRUCTION_CACHE_FLUSH |
 		  BRW_MI_GLOBAL_SNAPSHOT_RESET);
 	OUT_BATCH(MI_NOOP);
-	ADVANCE_BATCH();
 
 	/* brw_debug (scrn, "before base address modify"); */
-	if (IS_IGDNG(intel))
-		ATOMIC_BATCH(14);
-	else
-		ATOMIC_BATCH(12);
 	/* Match Mesa driver setup */
 	if (IS_G4X(intel) || IS_IGDNG(intel))
 		OUT_BATCH(NEW_PIPELINE_SELECT | PIPELINE_SELECT_3D);
@@ -834,9 +827,6 @@ i965_emit_video_setup(ScrnInfoPtr scrn, drm_intel_bo * bind_bo, int n_src_surf)
 	OUT_RELOC(intel->video.gen4_sip_kernel_bo,
 		  I915_GEM_DOMAIN_INSTRUCTION, 0, 0);
 
-	OUT_BATCH(MI_NOOP);
-	ADVANCE_BATCH();
-
 	/* brw_debug (scrn, "after base address modify"); */
 
 	if (IS_IGDNG(intel))
@@ -844,10 +834,6 @@ i965_emit_video_setup(ScrnInfoPtr scrn, drm_intel_bo * bind_bo, int n_src_surf)
 	else
 		pipe_ctl = BRW_PIPE_CONTROL_NOWRITE | BRW_PIPE_CONTROL_IS_FLUSH;
 
-	ATOMIC_BATCH(38);
-
-	OUT_BATCH(MI_NOOP);
-
 	/* Pipe control */
 	OUT_BATCH(BRW_PIPE_CONTROL | pipe_ctl | 2);
 	OUT_BATCH(0);		/* Destination address */
@@ -971,16 +957,14 @@ i965_emit_video_setup(ScrnInfoPtr scrn, drm_intel_bo * bind_bo, int n_src_surf)
 			   VE1_VFCOMPONENT_3_SHIFT) | (4 <<
 						       VE1_DESTINATION_ELEMENT_OFFSET_SHIFT));
 	}
-
-	OUT_BATCH(MI_NOOP);	/* pad to quadword */
-	ADVANCE_BATCH();
 }
 
 void
 I965DisplayVideoTextured(ScrnInfoPtr scrn,
 			 intel_adaptor_private *adaptor_priv, int id,
 			 RegionPtr dstRegion,
-			 short width, short height, int video_pitch,
+			 short width, short height,
+			 int video_pitch, int video_pitch2,
 			 short src_w, short src_h,
 			 short drw_w, short drw_h, PixmapPtr pixmap)
 {
@@ -1023,7 +1007,7 @@ I965DisplayVideoTextured(ScrnInfoPtr scrn,
 		src_surf_format = BRW_SURFACEFORMAT_R8_UNORM;
 		src_width[1] = src_width[0] = width;
 		src_height[1] = src_height[0] = height;
-		src_pitch[1] = src_pitch[0] = video_pitch * 2;
+		src_pitch[1] = src_pitch[0] = video_pitch2;
 		src_width[4] = src_width[5] = src_width[2] = src_width[3] =
 		    width / 2;
 		src_height[4] = src_height[5] = src_height[2] = src_height[3] =
@@ -1219,7 +1203,6 @@ I965DisplayVideoTextured(ScrnInfoPtr scrn,
 
 		i965_emit_video_setup(scrn, bind_bo, n_src_surf);
 
-		ATOMIC_BATCH(12);
 		/* Set up the pointer to our vertex buffer */
 		OUT_BATCH(BRW_3DSTATE_VERTEX_BUFFERS | 3);
 		/* four 32-bit floats per vertex */
@@ -1228,7 +1211,7 @@ I965DisplayVideoTextured(ScrnInfoPtr scrn,
 		OUT_RELOC(vb_bo, I915_GEM_DOMAIN_VERTEX, 0, 0);
 		if (IS_IGDNG(intel))
 			OUT_RELOC(vb_bo, I915_GEM_DOMAIN_VERTEX, 0,
-				  (vb_bo->offset + i) * 4);
+				  i * 4);
 		else
 			OUT_BATCH(3);	/* four corners to our rectangle */
 		OUT_BATCH(0);	/* reserved */
@@ -1241,7 +1224,6 @@ I965DisplayVideoTextured(ScrnInfoPtr scrn,
 		OUT_BATCH(0);	/* start instance location */
 		OUT_BATCH(0);	/* index buffer offset, ignored */
 		OUT_BATCH(MI_NOOP);
-		ADVANCE_BATCH();
 
 		intel_batch_end_atomic(scrn);
 
